@@ -13,30 +13,24 @@ from pathlib import Path    #
 import os
 from re import sub
 
-class Session:
-    """ Use an API account from a credentials file as described in https://developer.netfoundry.io/v2/guides/authentication/
-    Example credentials file:
-    {
-        "clientId": "3tcm6to3qqfu78juj9huppk9g3",
-        "password": "149a7ksfj3t5lstg0pesun69m1l4k91d6h8m779l43q0ekekr782",
-        "authenticationUrl": "https://netfoundry-production-xfjiye.auth.us-east-1.amazoncognito.com/oauth2/token"
-    }
+class Organization:
+    """ Default is to use the Organization of the caller's user or API account identity
+    :param: organization_id is optional string UUID of an alternative Organization
+    :param: organization_label is optional string `label` property of an alternative Organization
+    :param token: continue using a session with this optional token from an existing instance of Organization
+    :param credentials: optional alternative path to API account credentials file, default is ~/.netfoundry/credentials.json
+    :param proxy: optional HTTP proxy, e.g., http://localhost:8080
     """
 
-    def __init__(
-        self, 
+    def __init__(self, 
+        organization_id: str=None, 
+        organization_label: str=None,
         token=None, 
         credentials=None, 
         proxy=None):
-        """initialize with a reusable API client
-        :param token: optional temporary API bearer/session token
-        :param credentials: optional alternative path to API account credentials file
-        :param proxy: optional HTTP proxy, e.g., http://localhost:8080
-        The init function also gathers a few essential objects from the API and
-        stores them as attributes on the model
-        """
 
         # verify auth endpoint's server certificate if proxy is type SOCKS or None
+        self.proxy = proxy
         if proxy is None:
             self.proxies = dict()
             self.verify = True
@@ -57,31 +51,26 @@ class Session:
             self.token = os.environ['NETFOUNDRY_API_TOKEN']
 
         # if the token was found then extract the expiry
-        try: self.token
+        try: 
+            self.token
         except AttributeError: epoch = None
         else:
             claim = jwt.decode(jwt=self.token, algorithms=["RS256"], options={"verify_signature": False})
             # TODO: [MOP-13438] auto-renew token when near expiry (now+1hour in epoch seconds)
             expiry = claim['exp']
             epoch = time.time()
-#            print("DEBUG: found API token in env NETFOUNDRY_API_TOKEN")
 
-        # if no token or near expiry then use credentials to obtain a token
-        if epoch is not None and epoch < (expiry - 600):
-            # extract the API URL from the claim
-            self.audience = claim['scope'].replace('/ignore-scope','')
-            # e.g. https://gateway.production.netfoundry.io/
-#            print("DEBUG: using API token from env NETFOUNDRY_API_TOKEN")
+        # persist the credentials filename in instances so that it may be used to refresh the token
+        if credentials is not None:
+            self.credentials = credentials
+            os.environ['NETFOUNDRY_API_ACCOUNT'] = self.credentials
+        elif 'NETFOUNDRY_API_ACCOUNT' in os.environ:
+            self.credentials = os.environ['NETFOUNDRY_API_ACCOUNT']
         else:
-            # persist the credentials filename in instances so that it may be used to refresh the token
-            if credentials is not None:
-                self.credentials = credentials
-                os.environ['NETFOUNDRY_API_ACCOUNT'] = self.credentials
-            elif 'NETFOUNDRY_API_ACCOUNT' in os.environ:
-                self.credentials = os.environ['NETFOUNDRY_API_ACCOUNT']
-            else:
-                self.credentials = "credentials.json"
+            self.credentials = "credentials.json"
 
+        # if no token or near expiry (30 min) then use credentials to obtain a token
+        if epoch is None or epoch > (expiry - 600):
             # unless a valid path assume relative and search the default chain
             if not os.path.exists(self.credentials):
                 default_creds_chain = [
@@ -168,26 +157,93 @@ class Session:
                     )
                 )
 
-class Organization:
-    """ Use an organization
-    """
+        # learn about the environment from the token
+        try:
+            claim = jwt.decode(jwt=self.token, algorithms=["RS256"], options={"verify_signature": False})
+            iss = claim['iss']
+            if re.match(r'https://cognito-', iss):
+                self.environment = re.sub(r'https://gateway\.([^.]+)\.netfoundry\.io.*',r'\1',claim['scope'])
+            elif re.match(r'.*\.auth0\.com', iss):
+                self.environment = re.sub(r'https://netfoundry-([^.]+)\.auth0\.com.*',r'\1',claim['iss'])
+            # import q; q(claim)
+            # import epdb; epdb.serve()
+            self.audience = 'https://gateway.'+self.environment+'.netfoundry.io/'
+        except: raise
 
-    def __init__(self, Session):
-        self.session = Session
         # always resolve Network Groups so we can specify either name or ID when calling super()
         self.network_groups = self.get_network_groups_by_organization()
-        self.describe = self.get_organization()
-        self.label = self.describe['label']
+        # Obtain a session token and find own `.caller` identity and `.organizations`
+        self.caller = self.get_caller_identity()
+        if organization_id:
+            self.describe = self.get_organization(id=organization_id)
+        elif organization_label:
+            self.organizations_by_label = dict()
+            for org in self.get_organizations():
+                self.organizations_by_label[org['label']] = org['id']
+            self.describe = self.get_organization(id=self.organizations_by_label[organization_label])
+        else:
+            self.describe = self.get_organization(id=self.caller['organizationId'])
 
-    def get_organization(self):
-        """return the Organizations object (formerly "tenants")
+        self.label = self.describe['label']
+        self.id = self.describe['id']
+
+        
+    def get_caller_identity(self):
+        """return the caller's identity object
+        """
+        # try the API account endpoint first, then the endpoint for human, interactive users
+        request = {
+            "url": self.audience+'identity/v1/api-account-identities/self',
+            "proxies": self.proxies,
+            "verify": self.verify,
+            "headers": { "authorization": "Bearer " + self.token }
+        }
+        try:
+            response = requests.get(**request)
+            response_code = response.status_code
+        except:
+            raise
+
+        if response_code == requests.status_codes.codes.OK: # HTTP 200
+            try:
+                caller = json.loads(response.text)
+            except ValueError as e:
+                eprint('ERROR getting caller\'s API account identity from response document')
+                raise(e)
+        else:
+            try:
+                request["url"] = self.audience+'identity/v1/user-identities/self'
+                response = requests.get(**request)
+                response_code = response.status_code
+            except:
+                raise
+
+            if response_code == requests.status_codes.codes.OK: # HTTP 200
+                try:
+                    caller = json.loads(response.text)
+                except ValueError as e:
+                    eprint('ERROR getting caller\'s user identity from response document')
+                    raise(e)
+            else:
+                raise Exception(
+                    'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
+                        requests.status_codes._codes[response_code][0].upper(),
+                        response_code,
+                        response.text
+                    )
+                )
+
+        return(caller)
+
+    def get_organizations(self):
+        """return the list of Organizations (formerly "tenants")
         """
         try:
-            headers = { "authorization": "Bearer " + self.session.token }
+            headers = { "authorization": "Bearer " + self.token }
             response = requests.get(
-                self.session.audience+'identity/v1/organizations',
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'identity/v1/organizations',
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers
             )
             response_code = response.status_code
@@ -209,19 +265,50 @@ class Organization:
                 )
             )
 
-        return(organizations[0])
+        return(organizations)
 
+    def get_organization(self, id):
+        """return a single Organizations by ID
+        """
+        try:
+            headers = { "authorization": "Bearer " + self.token }
+            response = requests.get(
+                self.audience+'identity/v1/organizations/'+id,
+                proxies=self.proxies,
+                verify=self.verify,
+                headers=headers
+            )
+            response_code = response.status_code
+        except:
+            raise
+
+        if response_code == requests.status_codes.codes.OK: # HTTP 200
+            try:
+                organization = json.loads(response.text)
+            except ValueError as e:
+                eprint('ERROR getting Network Groups')
+                raise(e)
+        else:
+            raise Exception(
+                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
+                    requests.status_codes._codes[response_code][0].upper(),
+                    response_code,
+                    response.text
+                )
+            )
+
+        return(organization)
 
     def get_network_group(self,network_group_id):
         """describe a Network Group
         """
         try:
             # /network-groups/{id} returns a Network Group object
-            headers = { "authorization": "Bearer " + self.session.token }
+            headers = { "authorization": "Bearer " + self.token }
             response = requests.get(
-                self.session.audience+'rest/v1/network-groups/'+network_group_id,
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'rest/v1/network-groups/'+network_group_id,
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers
             )
             response_code = response.status_code
@@ -250,11 +337,11 @@ class Organization:
         """
         try:
             # /networks/{id} returns a Network object
-            headers = { "authorization": "Bearer " + self.session.token }
+            headers = { "authorization": "Bearer " + self.token }
             response = requests.get(
-                self.session.audience+'rest/v1/networks/'+network_id,
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'rest/v1/networks/'+network_id,
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers
             )
             response_code = response.status_code
@@ -280,14 +367,15 @@ class Organization:
 
     def get_network_groups_by_organization(self):
         """list Network Groups
+        TODO: filter by Organization when that capability is available
         """
         try:
             # /network-groups returns a list of dicts (Network Group objects)
-            headers = { "authorization": "Bearer " + self.session.token }
+            headers = { "authorization": "Bearer " + self.token }
             response = requests.get(
-                self.session.audience+'rest/v1/network-groups',
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'rest/v1/network-groups',
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers
             )
             response_code = response.status_code
@@ -318,11 +406,11 @@ class Organization:
 
         try:
             # returns a list of dicts (network objects)
-            headers = { "authorization": "Bearer " + self.session.token }
+            headers = { "authorization": "Bearer " + self.token }
             response = requests.get(
-                self.session.audience+'core/v2/networks',
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'core/v2/networks',
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers
             )
             response_code = response.status_code
@@ -352,15 +440,15 @@ class Organization:
         """
         try:
             headers = { 
-                "authorization": "Bearer " + self.session.token 
+                "authorization": "Bearer " + self.token 
             }
             params = {
                 "findByNetworkGroupId": network_group_id
             }
             response = requests.get(
-                self.session.audience+'core/v2/networks',
-                proxies=self.session.proxies,
-                verify=self.session.verify,
+                self.audience+'core/v2/networks',
+                proxies=self.proxies,
+                verify=self.verify,
                 headers=headers,
                 params=params
             )
@@ -390,7 +478,7 @@ class Organization:
 class NetworkGroup:
     """use a Network Group by name or ID or the first group in the organization
     """
-    def __init__(self, Organization, network_group_id=None, network_group_name=None):
+    def __init__(self, Organization: object, network_group_id: str=None, network_group_name: str=None):
         if network_group_id:
             self.network_group_id = network_group_id
             self.network_group_name = [ ng['organizationShortName'] for ng in Organization.network_groups if ng['id'] == network_group_id ][0]
@@ -416,25 +504,16 @@ class NetworkGroup:
         else:
             raise Exception("ERROR: need at least one Network Group in organization")
 
-        self.session = Organization.session
+        self.session = Organization
         self.describe = Organization.get_network_group(self.network_group_id)
         self.id = self.network_group_id
         self.name = self.network_group_name
         self.vanity = Organization.label.lower()
 
-        # learn about the environment from the token and predict the web console URL
-        try:
-            claim = jwt.decode(jwt=self.session.token, algorithms=["RS256"], options={"verify_signature": False})
-            iss = claim['iss']
-            if re.match('.*cognito.*', iss):
-                self.environment = re.sub(r'https://gateway\.([^.]+)\.netfoundry\.io.*',r'\1',claim['scope'])
-            elif re.match('auth0', iss):
-                self.environment = re.sub(r'https://netfoundry-([^.]+)\.auth0\.com.*',r'\1',claim['iss'])
-            if self.environment == "production":
-                self.nfconsole = "https://{vanity}.nfconsole.io".format(vanity=self.vanity)
-            else:
-                self.nfconsole = "https://{vanity}.{env}-nfconsole.io".format(vanity=self.vanity, env=self.environment)
-        except: raise
+        if self.session.environment == "production":
+            self.nfconsole = "https://{vanity}.nfconsole.io".format(vanity=self.vanity)
+        else:
+            self.nfconsole = "https://{vanity}.{env}-nfconsole.io".format(vanity=self.vanity, env=self.session.environment)
 
         # an attribute that is a dict for resolving network UUIDs by name
         self.networks_by_name = dict()
@@ -629,18 +708,18 @@ class NetworkGroup:
 class Network:
     """describe and use a Network
     """
-    def __init__(self, Session, network_id=None, network_name=None):
+    def __init__(self, NetworkGroup: object, network_id: str=None, network_name: str=None):
         """
-        :param token: required bearer token for this session
+        :param NetworkGroup: required object of the parent Network Group of this Network
         :param network_name: optional name of the network to describe and use
         :param network_id: optional UUID of the network to describe and use
         """
-        self.session = Session
+        self.session = NetworkGroup.session
 
         if network_id:
             self.describe = self.get_network_by_id(network_id)
         elif network_name:
-            self.describe = self.get_network_by_name(network_name)
+            self.describe = self.get_network_by_name(name=network_name,group=NetworkGroup.id)
         else:
             raise Exception("ERROR: need one of network_id or network_name")
 
@@ -787,7 +866,7 @@ class Network:
             params = {
                 "networkId": self.id,
                 "page": 0,
-                "size": 10,
+                "size": 100,
                 "sort": "name,asc"
             }
             if type == "services": 
@@ -910,6 +989,8 @@ class Network:
 
         # attempt to update if there's at least one difference between the current resource and the submitted patch
         if len(pruned_patch.keys()) > 0:
+            if not "name" in pruned_patch.keys():
+                pruned_patch["name"] = before_resource["name"]
             try:
                 after_response = requests.patch(
                     patch['_links']['self']['href'],
@@ -978,8 +1059,12 @@ class Network:
             )
         return(resource)
 
-    def create_endpoint(self, name, attributes=[]):
+    def create_endpoint(self, name: str, attributes: list=[], session_identity: str=None):
         """create an Endpoint
+        :param: name is required string on which to key future operations for this Endpoint
+        :param: attributes is an optional list of Endpoint roles of which this Endpoint is a member
+        :param: session_identity is optional string UUID of the identity in the NF Organization for
+                which a concurrent web console session is required to activate this Endpoint
         """
         try:
             headers = { 
@@ -994,6 +1079,10 @@ class Network:
                 "attributes": attributes,
                 "enrollmentMethod": { "ott": True }
             }
+
+            if session_identity:
+                body['sessionIdentityId'] = session_identity
+
             response = requests.post(
                 self.session.audience+'core/v2/endpoints',
                 proxies=self.session.proxies,
@@ -1179,7 +1268,7 @@ class Network:
                                 else: bind_endpoints += ['@'+endpoint_name] 
                             else:
                                 try:
-                                    name_lookup = self.get_resources(type="endpoints",id=endpoint)
+                                    name_lookup = self.get_resource(type="endpoint",id=endpoint)
                                     endpoint_name = name_lookup['name']
                                 except Exception as e:
                                     raise Exception('ERROR: Failed to find exactly one hosting Endpoint with ID "{}". Caught exception: {}'.format(endpoint, e))
@@ -1288,9 +1377,10 @@ class Network:
 
         return(app_wan)
 
-    def get_network_by_name(self,name):
+    def get_network_by_name(self,name: str,group: str=None):
         """return exactly one network object
-            :name required name of the NF network may contain quoted whitespace
+        :param: name required name of the NF network may contain quoted whitespace
+        :param: group optional string UUID to limit results by Network Group ID
         """
         try:
             headers = { 
@@ -1299,6 +1389,9 @@ class Network:
             params = {
                 "findByName": name
             }
+            if group is not None:
+                params['findByNetworkGroupId'] = group
+
             response = requests.get(
                 self.session.audience+'core/v2/networks',
                 proxies=self.session.proxies,
@@ -1586,7 +1679,7 @@ class Network:
             params = dict()
             if not type == "network":
                 params["networkId"] = self.id
-            elif type == "service": 
+            if type == "service": 
                 params["beta"] = ''
 
             response = requests.get(
@@ -1641,15 +1734,18 @@ class Network:
         :param id: required entity UUID to delete
         :param wait: optional seconds to wait for entity destruction
         """
+#        import epdb; epdb.serve()
         try:
             headers = { "authorization": "Bearer " + self.session.token }
             entity_url = self.session.audience+'core/v2/networks/'+self.id
-            expect = requests.status_codes.codes.ACCEPTED
+            expected_responses = [
+                requests.status_codes.codes.ACCEPTED,
+                requests.status_codes.codes.OK
+            ]
             if not type == 'network':
                 if id is None:
                     raise Exception("ERROR: need entity UUID to delete")
                 entity_url = self.session.audience+'core/v2/'+type+'s/'+id
-                expect = requests.status_codes.codes.OK
             eprint("WARN: deleting {:s}".format(entity_url))
             # TODO: remove "beta" when legacy Services API is decommissioned in favor of Platform Services API
             # results in HTMLv5-compliant URL param singleton with empty string value like ?beta= to invoke the Platform Services API
@@ -1668,7 +1764,7 @@ class Network:
         except:
             raise
 
-        if not response_code == expect:
+        if not response_code in expected_responses:
             raise Exception(
                 'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                     requests.status_codes._codes[response_code][0].upper(),
@@ -1763,6 +1859,10 @@ RESOURCES = {
     'services': {
         'embedded': "serviceList",
         'expect': "ACCEPTED"
+    },
+    'posture-checks': {
+        'embedded': "postureCheckList",
+        'expect': "ACCEPTED"
     }
 }
 
